from ..Template import *
from .. import Error
import datetime


def _data_dict(fields: tuple, datas: list) -> list:
    res = []
    keys = range(len(fields))
    for v in datas:
        tmp = {}
        for k in keys:
            tmp.update({fields[k]: str(v[k]) if isinstance(v[k], datetime.datetime) else v[k]})
        res.append(tmp)
    return res

def _data_table(fields: tuple, datas: list) -> dict:
    res = {
        'header': list(fields),
        'data': [[str(d) if isinstance(d, datetime.datetime) else d for d in i] for i in datas],
    }
    return res

def _format_field(field):
    # field = []
    res = []
    if isinstance(field, dict):
        for k, v in dict(field).items():
            tmp = k.split('__', 1)
            if len(tmp) == 2:
                res.append(f'{tmp[0]}.`{tmp[1]}` AS `{v}`' if v else f'{tmp[0]}.`{tmp[1]}`')
            else:
                res.append(f'`{k}` AS `{v}`' if v else f'`{k}`')
    elif isinstance(field, (tuple, list)):
        res = list(field)
    else:
        raise Error.ParamsError(400001)
    return res

def _filter_format(filter_data:dict) -> str:
    # 检查
    if set(filter_data.keys()) >= set(['symbol', 'field', 'value']):
        # check symbol
        field, val = f"`{filter_data.get('field')}`", filter_data.get('value')
        symbol = filter_data.get('symbol')
    
        symbol_list = ['=', '<=', '>=', '!=', '<>', '<', '>', 'IN', 'LIKE', 'BETWEEN', 'ISNULL']
        if symbol.upper() in symbol_list:
            result = ''
            # 表 as 名称替换 field
            if filter_data.get('table_as'):
                field = f"{filter_data.get('table_as')}.{field}"
            # symbol 格式化
            if symbol.upper() == 'IN':
                if filter_data.get('not'):
                    symbol = 'NOT IN'
                result = f"{field} {symbol} ({','.join([str(i) for i in val])})"
            elif symbol.upper() == 'LIKE':
                if filter_data.get('not'):
                    symbol = 'NOT LIKE'
                result = f"{field} {symbol} {str(val)}"
            elif symbol.upper() == 'ISNULL':
                symbol = 'IS NULL'
                if filter_data.get('not'):
                    symbol = 'IS NOT NULL'
                result = f'{field} {symbol}'
            elif symbol.upper() == 'BETWEEN':
                if filter_data.get('not'):
                    symbol = 'NOT BETWEEN'
                result = f"{field} {symbol} {val.get('start')} AND {val.get('end')}"
                if filter_data.get('not'):
                    symbol = 'NOT LIKE'
                result = f'{field} {symbol} {str(val)}'
            else:
                result = f'{field} {symbol} {val}'
                if filter_data.get('not'):
                    result = f'NOT {result}'
            # 关系
            if factor := filter_data.get('factor'):
                if factor.upper() in ['AND', 'OR']:
                    result = f'{factor.upper()} {result}'
                else:
                    raise Error.ParamsError(400004)
            return result
        else:
            raise Error.ParamsError(400004)
    else:
        raise Error.ParamsError(400003, 'Params miss in [symbol, field, value]')

class DB(TDB):

    def __init__(self):
        TDB.__init__(self)
        self.db_type = 'mysql'
        self.db_info = {
            'database': None,
            'user': None,
            'passwd': None,
            'host': 'localhost',
            'port': 3306
        }
        import mysql.connector
        self.device = mysql.connector
        self.__init_conf = False
        self.__db = None

    def set_conf(self, **db_info):
        # up data key
        upkey = {
            'password': 'passwd'
        }
        for k, v in upkey.items():
            if db_info.get(k):
                db_info.update({v: db_info.pop(k)})
        need_key = [k for k, v in self.db_info.items() if v == None]
        if not set(db_info.keys()) >= set(need_key):
            raise Error.DatabaseConfError(100001, set(need_key) - set(db_info))
        # sign init conf
        self.db_info.update(db_info)
        self.__init_conf = True

    def run(self, sql):
        if sql != None and self.__connect:
            __type = sql.get_type()
            value = sql.get_sql()
            # 游标
            dbcursor = self.__db.cursor()
            if __type == 'query':
                try:
                    dbcursor.execute(value)
                except Exception as e:
                    print(e)
                else:
                    sql.data = {
                        'field': dbcursor.column_names,
                        'data': dbcursor.fetchall()
                    }
            else:
                if isinstance(value, list):
                    for i in value:
                        try:
                            dbcursor.execute(i)
                        except Exception as e:
                            print(e)
                else:
                    try:
                        dbcursor.execute(value)
                    except Exception as e:
                        print(e)

    def commit(self):
        try:
            self.__db.commit()
        except Exception as e:
            print(e)
            raise Error.CommitError(100022)

    def rollback(self):
        try:
            self.__db.rollback()
        except Exception as e:
            raise Error.RollbackError(100023)

    def connect(self):
        # 检查 配置 初始化 通过
        if self.__init_conf:
            try:
                self.__db = self.device.connect(**self.db_info)
                self.__connect = True
            except self.device.errors.InterfaceError as e:
                raise Error.CommitError(100011, e)
            except Exception as e:
                raise Error.ConnectError(100010, e)
        else: raise Error.UseError(200001, 'conf not init, please use "set_conf" method.')

    def close(self):
        if self.__db:
            self.__db.close()
        else:
            raise Error.UseError(200001)

class Insert(TInsert):

    def __init__(self, table, data):
        TInsert.__init__(self, table, data)
        if isinstance(data, dict):
            self.fields = list(data.keys())
        elif isinstance(data, list) and  len(data) > 0:
            if isinstance(data[0], dict):
                self.fields = list(data[0].keys())
            else:
                Error.ParamsError(400001)
        else:
            Error.ParamsError(400001)
        self.__base = ['INSERT INTO', 'VALUES']
        self.val_max = 5
        self.values = []

    def data_decode(self, data = None):
        if isinstance(self.data, dict) or data:
            data:dict = data if data else self.data
            data_list = [repr(data.get(i)) if i in data.keys() else 'NULL' for i in self.fields]
            self.values.append(f"({','.join(data_list)})")
        else:
            if isinstance(self.data, list):
                for i in self.data:
                    self.data_decode(i)
            else:
                raise Error.ParamsError(400001)

    def format_sql(self):
        if not self.values:
            self.data_decode()
        res = []
        field = ','.join([f'`{i}`' for i in self.fields])
        for i in range(0, len(self.values), self.val_max):
            tmp = self.values[i: i * self.val_max + self.val_max]
            res.append(f"{self.__base[0]} {self.table} ({field}) {self.__base[1]} {','.join(tmp)}")
        self.set_sql(res)

class Update(TUpdate):

    def __init__(self, table, data):
        TUpdate.__init__(self, table, data)
        self.__base = ['UPDATE', 'SET']

    def data_decode(self):
        if hasattr(self, 'data'):
            if isinstance(self.data, dict):
                self.value = [f'`{k}` = {repr(v)}' for k, v in self.data.items()]
            else:
                raise Error.ParamsError(400001)
        else:
            raise Error.UseError(200002)

    def format_sql(self):
        self.data_decode()
        str_data = ','.join(self.value)
        sql = f'{self.__base[0]} {self.table} {self.__base[1]} {str_data}'
        if self.where:
            sql = f'{sql} {self.where}'
        self.set_sql(sql)

    def set_where(self, *filters):
        if len(filters) > 0:
            self.where = []
            for i in filters:
                self.where.append(_filter_format(i))
            self.where = f'WHERE {" ".join(self.where)}'
        else:
            raise Error.ParamsError(400002)

class Delete(TDelete):

    def __init__(self, table):
        TDelete.__init__(self, table)
        self.__base = ['DELETE FROM']

    def format_sql(self):
        sql = f'{self.__base[0]} {self.table}'
        if self.where:
            sql = f'{sql} {self.where}'
        self.set_sql(sql)

    def set_where(self, *filters):
        if len(filters) > 0:
            self.where = []
            for i in filters:
                self.where.append(_filter_format(i))
            self.where = f'WHERE {" ".join(self.where)}'
        else:
            raise Error.ParamsError(400002)

class Select(TSelect):

    def __init__(self, **kwargs):
        TSelect.__init__(self, **kwargs)
        self.__base = ['SELECT', 'FROM']
        if kwargs.get('fields'):
            if isinstance(kwargs.get('fields'), (list, tuple, dict)):
                self.fields = _format_field(kwargs.get('fields'))
        else:
            self.fields = '*'

    def set_fields(self, *fields):
        if len(fields) < 1:
            raise Error.ParamsError(400002)
        for i in fields:
            self.fields = _format_field(i)


    # 设置条件
    def set_where(self, *filters):
        if len(filters) > 0:
            self.where = []
            for i in filters:
                self.where.append(_filter_format(i))
            self.where = f'WHERE {" ".join(self.where)}'
        else:
            raise Error.ParamsError(400002)

    # 设置分组
    def set_group(self, *fields):
        '''
        Set up groups
            *fields: table field
                [field] or [table]__[field]
        '''
        if fields:
            f_list = []
            for i in fields:
                tmp = i.split('__', 1)
                f_list.append(f'{tmp[0]}.`{tmp[1]}`' if len(tmp) == 2 else f'`{i}`')
            if f_list:
                self.group = 'GROUP BY {}'.format(','.join(f_list))
        else:
            raise Error.ParamMiss('fields')

    # 设置排序
    def set_order(self, **fields):
        """设置排序

        Raises:
            Error.ParamsError: _description_
        """
        if fields:
            f_list = []
            for k, v in fields.items():
                tmp = k.split('__', 1)
                if len(tmp) == 2:
                    f_list.append(f'{tmp[0]}.`{tmp[1]}` DESC' if v else f'{tmp[0]}.`{tmp[1]}`')
                else:
                    f_list.append(f'`{k}` DESC' if v else f'`{k}`')
            if f_list:
                self.order = 'ORDER BY {}'.format(','.join(f_list))
        else:
            raise Error.ParamsError(400002)

    # 设置分页
    def set_limit(self, row, num = None):
        if isinstance(row, int):
            self.limit = f'LIMIT {row}'
        else:
            raise Error.ParamsError(400001)
        if isinstance(num, int):
            self.limit = f'{self.limit}, {num}'
        else:
            if num != None:
                raise Error.ParamsError(400001)

    # 设置链表查询
    def set_join(self, *join_obj) -> str:
        '''
        set join sql object
            *join_obj: Join table object
                {
                    'keyword': ['join', 'left', 'right'],
                    'sql': str(sql),
                    'table': str(table_name),
                    'table_as': str(t),
                    'col_key': str(col_name),
                    'col_value': str(table.col_name),
                }
        '''
        tmp = []
        if join_obj:
            for obj in join_obj:
                if obj.get('keyword').lower() == 'left':
                    obj['keyword'] = 'LEFT JOIN'
                elif obj.get('keyword').lower() == 'right':
                    obj['keyword'] = 'RIGHT JOIN'
                elif obj.get('keyword').lower() == 'join':
                    obj['keyword'] = 'JOIN'
                else:
                    raise Error.ParamsError(400003)
                if 'sql' in join_obj:
                    tmp.append(f"{str(obj['keyword'])} ({str(obj['sql'])}) {str(obj['table_as'])} ON {str(obj['table_as'])}.`{obj['col_key']}` = {str(obj['col_value'])}")
                else:
                    tmp.append(f"{str(obj['keyword'])} {str(obj['table'])} {str(obj['table_as'])} ON {str(obj['table_as'])}.`{obj['col_key']}` = {str(obj['col_value'])}")
        else:
            raise Error.ParamsError(400002)
        if tmp:
            self.join = ' '.join(tmp)

    def format_sql(self):
        sql = ''
        if self.table and self.fields:
            if isinstance(self.table, dict):
                self.table = [f'{k} {v}' for k, v in self.table.items()]
            elif isinstance(self.table, str):
                self.table = [self.table]
            table = ','.join(self.table)
            field = ','.join(self.fields)
            sql = f'{self.__base[0]} {field} {self.__base[1]} {table}'
            if self.join:
                sql = f'{sql} {self.join}'
            if self.where:
                sql = f'{sql} {self.where}'
            if self.group:
                sql = f'{sql} {self.group}'
            if self.order:
                sql = f'{sql} {self.order}'
            if self.limit:
                sql = f'{sql} {self.limit}'
            self.set_sql(sql)
        else:
            raise Error.ParamsError(400002)

    def format_data(self, t = 'dict') -> list|dict:
        """格式化数据结果

        Args:
            t (str, optional): 模板. Defaults to 'dict'.

        Raises:
            Error.UseError: 参数异常 对象必须有 data 属性

        Returns:
            list|dict: 结果
        """
        # 检查数据是否存在
        if not hasattr(self, 'data'):
            raise Error.UseError(200001)
        if 'field' in self.data.keys() and 'data' in self.data.keys():
            if t == 'dict':
                self.res = _data_dict(self.data.get('field'), self.data.get('data'))
            elif t == 'table':
                self.res = _data_table(self.data.get('field'), self.data.get('data'))
            return self.res

__all__ = [
    'Insert',
    'Select',
    'Update',
    'Delete',
]


# Error
# 1----- 数据库相关异常
# 2----- 类使用方法异常
# 3----- 方法运行异常
# 4----- 输入异常
# 5----- 输出处理异常

message = {
    # 未知异常
    000000: 'Unknown exception.',
    # 数据库相关异常
    # 数据库 配置参数 缺少
    100001: 'Missing database conf params.',
    # 数据库 配置参数异常
    100002: 'Database conf params error.',
    # 数据库 连接失败
    100010: 'Database connection failed.',
    # 数据库 无法连接mysql服务
    100011: 'Can\'t connect to MySQL server.',
    # 数据库 连接超时
    100012: 'Database connection timeout.',
    # 执行 sql 异常
    100020: 'Run SQL Exception.',
    # 执行 sql 超时
    100021: 'SQL execution timeout.',
    # 提交 错误
    100022: 'Submission failed.',
    # 回滚 错误
    100023: 'Rollback failed.',

    # 类使用方法异常
    # 对象 不能使用此方法
    200001: 'This object cannot use this method.',
    # 对象没有这个属性
    200002: 'The object dost not have the attribute.',
  
    # 方法运行异常
    # 模板格式化错误
    300001: 'Template format error.',
    # 输入异常
    
    # 参数异常
    400001: 'Params type error.',
    400002: 'Params miss.',
    400003: 'Params format error.',
    400004: 'Params value error.',
    
    # 输出异常
    # 结果处理异常
    500000: 'Result processing error.',
}

class BaseError(Exception):

    def __init__(self, code = 000000, /, *args):
        # 检查 code 是否存在
        self.code = code if code in message.keys() else 000000
        self.msg = message.get(self.code)
        self.res = Exception(*args)
        self.__str__()
    
    def __str__(self):
        return f'[{self.code:0>6}] [{self.__class__.__name__}] -> {self.msg}\nErrorMessage: {self.res}'

# test
class TestError(BaseError):...

# 数据库相关异常
class DatabaseConfError(BaseError):...
class ConnectError(BaseError):...
class RunSqlError(BaseError):...
class RunSqlTimeoutError(BaseError):...
class CommitError(BaseError):...
class RollbackError(BaseError):...
# 代码错误
class UseError(BaseError):...
class ParamsError(BaseError):...
class FormatError(BaseError):...

__all__ = [
    'ConnectError'
    'UseError'
    'TestError'
    'ParamsError'
    'FormatError'
]



from .. import Error
from .base import DB, Sql
from .base import (
    Insert as JI,
    Delete as JD,
    Update as JU,
    Select as JS,
)

class TDB(DB):...

class TInsert(JI, Sql):

    def __init__(self, table, data:list|dict):
        JI.__init__(self)
        Sql.__init__(self)
        self.set_type('affairs')
        self.table = table
        self.data = data

    def data_decode(self): raise NotImplementedError
    
    def format_sql(self): raise NotImplementedError

class TUpdate(JU, Sql):

    def __init__(self, table, data:dict):
        JU.__init__(self)
        Sql.__init__(self)
        self.set_type('affairs')
        self.table = table
        self.data = data

    def data_decode(self): raise NotImplementedError

    def format_sql(self): raise NotImplementedError

class TDelete(JD, Sql):

    def __init__(self, table):
        JD.__init__(self)
        Sql.__init__(self)
        self.set_type('affairs')
        self.table = table

    def format_sql(self): raise NotImplementedError

class TSelect(JS, Sql):

    def __init__(self, /, **kwargs):
        JS.__init__(self)
        Sql.__init__(self)
        self.set_type('query')
        if kwargs.get('table'):
            self.table = kwargs.get('table')

    def format_sql(self): raise NotImplementedError

    def format_data(self): raise NotImplementedError

__all__ = [
    'TDB',
    'TInsert',
    'TUpdate',
    'TDelete',
    'TSelect',
]



import importlib
import functools
from . import db

__version__ = '1.0.0'

__cls__ = db.__all__


def SeniorDB(t):
    if t in __cls__:
        __db = getattr(db, t).DB
    else:
        __db = importlib.import_module(name = t).DB
    
    class SeniorDB(__db):...
    return SeniorDB()


def SeniorSQL(o: SeniorDB):
    t = o.db_type
    if t in __cls__:
        pack = getattr(db, t)
    else:
        pack = importlib.import_module(name = t)

    class SeniorSQL(*[getattr(pack, i) for i in pack.__all__]):
        
        def __init__(self):...

        def Add(self, *args, **kwargs):
            pack.Insert.__init__(self, *args, **kwargs)
            self.func = 'Insert'

        def Del(self, *args, **kwargs):
            pack.Delete.__init__(self, *args, **kwargs)
            self.func = 'Delete'

        def Up(self, *args, **kwargs):
            pack.Update.__init__(self, *args, **kwargs)
            self.func = 'Update'
        
        def Sel(self, *args, **kwargs):
            pack.Select.__init__(self, *args, **kwargs)
            self.func = 'Select'

        def __run_class_func(func):
            @functools.wraps(func)
            def f(self, *args, **kwargs):
                return getattr(getattr(pack, self.func), func.__name__)(self, *args, **kwargs)
            return f
        
        @__run_class_func
        def set_join(self, *args, **kwargs):...

        @__run_class_func
        def set_where(self, *args, **kwargs):...

        @__run_class_func
        def set_group(self, *args, **kwargs):...
        
        @__run_class_func
        def set_order(self, *args, **kwargs):...

        @__run_class_func
        def set_limit(self, *args, **kwargs):...

        @__run_class_func
        def data_decode(self, *args, **kwargs):...

        @__run_class_func
        def format_sql(self, *args, **kwargs):
            """格式化 SQL 语句
            """
            pass

        @__run_class_func
        def format_data(self, *args, **kwargs) -> list|dict:
            """格式化数据

            Returns:
                list|dict: 结果
            """
            pass

    return SeniorSQL()



import re
from .. import Error

def temp_format(o, t: str, **kwargs):
    res = t
    # check
    obj_list = re.findall(r'[a-z]+?(?=\[)', t)
    if sum([not hasattr(o, i) for i in obj_list]) > 0:
        raise Error.FormatError(200002, [i for i in obj_list if not hasattr(o, i)])
    for i in t.split():
        k = re.search(r'[a-z]+?\b', i).group(0)
        if '*' in i:
            v = kwargs.get(k)
            join_str = i[i.index(':') + 1] if ':' in i else ','
            res = res.replace(i[i.find(k): i.index(':') + 2], join_str.join(v))
        else:
            v = getattr(o, k)
            if '[]' in i:
                res = res.replace(i, v)
            else:
                ki = re.search(r'\d+', i).group(0)
                res = res.replace(i, v[int(ki)])
    return res

def filter_format(func, check: str, **kwargs) -> str:
    # template 't'
    # 变量对照检查
    t = '{_logic} {_not} {_t_as:.}`{field}` {symbl} {value}'
    print(t)
    return 

"""
基本 模型

DB 数据库管理

Sql 基本 sql 方法对象

Tab 表模型
针对表模型进行
增 删 改 查 扩展表模型 属性

"""

class DB:
    """DB 数据库对象
    """

    def __init__(self):
        self.db_type = None
        self.db_info = None
        self.__db = None

    def set_conf(self):
        raise NotImplementedError

    def get_conf(self):
        return self.db_info

    def connect(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError


class Sql:
    """Sql 数据类型
    """

    def __init__(self):
        # type: query/affairs
        self.__type = None
        self.__value = None

    def get_sql(self):
        return self.__value
    
    def set_sql(self, v):
        self.__value = v
    
    def get_type(self):
        return self.__type
    
    def set_type(self, t):
        self.__type = t


class Tab:

    def __init__(self):
        self.table = None

    def set_where(self):
        # check 当对象没有该属性无法使用方法
        if not hasattr(self, 'where'):
            raise Error.UseError(200001)
    
    def get_where(self):
        # check 当对象没有该属性无法获取
        if not hasattr(self, 'where'):
            raise Error.UseError(200002)

class Insert(Tab):

    def __init__(self):
        Tab.__init__(self)
        # 字段列表
        self.fields = []
        # 数据集
        self.data = None

class Delete(Tab):

    def __init__(self):
        Tab.__init__(self)
        self.where = None
    
    def get_where(self):
        Tab.get_where(self)
        return self.where

class Update(Tab):

    def __init__(self):
        Tab.__init__(self)
        self.where = None
    
    def get_where(self):
        Tab.get_where(self)
        return self.where

class Select(Tab):

    def __init__(self):
        Tab.__init__(self)
        self.fields = []
        self.join = None
        self.where = None
        self.group = None
        self.order = None
        self.limit = None
